################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################

"""
This module provide some tools
"""
from __future__ import absolute_import
from builtins import str, bytes
from future.utils import iteritems, iterkeys, itervalues
from baidubce import compat

import os
import re
import datetime
import hashlib
import base64
import string
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import baidubce
from baidubce.http import http_headers

import subprocess
import uuid
import platform
import psutil
import os
import socket
import getpass

from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend




def get_canonical_time(timestamp=0):
    """
    Get cannonical time.

    :type timestamp: int
    :param timestamp: None
    =======================
    :return:
        **string of canonical_time**
    """
    if timestamp == 0:
        utctime = datetime.datetime.utcnow()
    else:
        utctime = datetime.datetime.utcfromtimestamp(timestamp)
    return b"%04d-%02d-%02dT%02d:%02d:%02dZ" % (
        utctime.year, utctime.month, utctime.day,
        utctime.hour, utctime.minute, utctime.second)


def get_ip_address():
    """
    Get the ip address of the migrate source

    :return:
        **string**
    """
    p = subprocess.Popen("hostname -I", shell=True, stdout=subprocess.PIPE)
    data = p.stdout.read()
    data = str(data, encoding='UTF-8')
    ip_list = data.split(' ')
    if "\n" in ip_list:
        ip_list.remove("\n")
    return ip_list[0]


def get_mac_address():
    """
    Get the MAC address of the migrate source

    :return
    :rtype string
    """
    mac = uuid.UUID(int = uuid.getnode()).hex[-12:]
    return ":".join([mac[e: e + 2] for e in range(0, 11, 2)])


def get_os_info():
    """
    Get the os information of the migrate source

    :return
    :rtype string
    """
    return platform.platform()


def get_ssh_pub_key():
    """
    Get ssh public key. If not exist, create it and return

    :return
    :rtype string
    """
    home_path = os.environ['HOME']
    ssh_pub_key_path = home_path + "/.ssh/id_rsa.pub"
    ssh_pub_key = None
    if os.path.exists(ssh_pub_key_path):
        with open(ssh_pub_key_path, 'r') as f:
            ssh_pub_key = f.read()
    else:
        key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048
        )
        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption())
        public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH
        )
        ssh_private_key_path = home_path + "/.ssh/id_rsa"
        private_key = private_key.decode()
        public_key = public_key.decode()
        public_key = public_key + " " + getpass.getuser() + "@" + socket.gethostname() + "\n"

        with open(ssh_private_key_path, 'w') as f:
            f.write(private_key)
        with open(ssh_pub_key_path, 'w') as f:
            f.write(public_key)
        ssh_pub_key = public_key
    return ssh_pub_key


def get_disk_info():
    """
    Get the system disk information and the data disk information of the migrate source.

    :return:
    :rtype dict, list
    """
    partitions = psutil.disk_partitions()
    system_disk_info = {}
    data_disk_info = []
    for i in range(len(partitions)):
        if partitions[i].mountpoint == "/":
            system_disk_info["total_size"] = psutil.disk_usage(partitions[i].mountpoint).total
            system_disk_info["used_size"] = psutil.disk_usage(partitions[i].mountpoint).used
        else:
            tmp = {}
            tmp["total_size"] = psutil.disk_usage(partitions[i].mountpoint).total
            tmp["used_size"] = psutil.disk_usage(partitions[i].mountpoint).used
            data_disk_info.append(tmp)

    return system_disk_info, data_disk_info


def get_disk_size():
    """
    Get the size of system disk and data disks

    :return
    :rtype int, list
    """
    partitions = psutil.disk_partitions()
    system_disk_size = 0
    data_disks_size = []
    for i in range(len(partitions)):
        if partitions[i].mountpoint == "/":
            system_disk_size = psutil.disk_usage(partitions[i].mountpoint).total
        else:
            data_disks_size.append(psutil.disk_usage(partitions[i].mountpoint).total)
    return system_disk_size, data_disks_size


def get_partition_info():
    """
    Get the system partition information of the migrate source.

    :return:
    :rtype list
    """
    partitions = psutil.disk_partitions()
    partition_info = []
    for i in range(len(partitions)):
        device = partitions[i].device
        size = psutil.disk_usage(partitions[i].mountpoint).total
        file_system = partitions[i].fstype
        mount_point = partitions[i].mountpoint
        partition_info.append(
            {
                "device": device,
                "size": size,
                "file_system": file_system,
                "mount_point": mount_point
            }
        )

    return partition_info


def get_cpu():
    """
    Get the number of CPU cores of the migration source

    :return:
    :rtype list
    """
    return os.cpu_count()


def get_memory():
    """
    Get the memory of the migration source

    :return:
    :rtype list
    """
    return psutil.virtual_memory().total


def convert_to_standard_string(input_string):
    """
    Encode a string to utf-8.

    :type input_string: string
    :param input_string: None
    =======================
    :return:
        **string**
    """
    #if isinstance(input_string, str):
    #    return input_string.encode(baidubce.DEFAULT_ENCODING)
    #elif isinstance(input_string, bytes):
    #    return input_string
    #else:
    #    return str(input_string).encode("utf-8")
    return compat.convert_to_bytes(input_string)


def _get_normalized_char_list():
    """"
    :return:
        **ASCII string**
    """
    ret = ['%%%02X' % i for i in range(256)]
    for ch in string.ascii_letters + string.digits + '.~-_':
        ret[ord(ch)] = ch
    if isinstance(ret[0], str):
        ret = [s.encode("utf-8") for s in ret]
    return ret
_NORMALIZED_CHAR_LIST = _get_normalized_char_list()


def normalize_string(in_str, encoding_slash=True):
    """
    Encode in_str.
    When encoding_slash is True, don't encode skip_chars, vice versa.

    :type in_str: string
    :param in_str: None

    :type encoding_slash: Bool
    :param encoding_slash: None
    ===============================
    :return:
        **ASCII  string**
    """
    tmp = []
    for ch in convert_to_standard_string(in_str):
        # on python3, ch is int type
        sep = ''
        index = -1
        if isinstance(ch, int):
            # on py3
            sep = chr(ch).encode("utf-8")
            index = ch
        else:
            sep = ch
            index = ord(ch)
        if sep == b'/' and not encoding_slash:
            tmp.append(b'/')
        else:
            tmp.append(_NORMALIZED_CHAR_LIST[index])
    return (b'').join(tmp)



_first_cap_regex = re.compile('(.)([A-Z][a-z]+)')
_number_cap_regex = re.compile('([a-z])([0-9]{2,})')
_end_cap_regex = re.compile('([a-z0-9])([A-Z])')


def pythonize_name(name):
    """Convert camel case to a "pythonic" name.
    Examples::
        pythonize_name('CamelCase') -> 'camel_case'
        pythonize_name('already_pythonized') -> 'already_pythonized'
        pythonize_name('HTTPRequest') -> 'http_request'
        pythonize_name('HTTPStatus200Ok') -> 'http_status_200_ok'
        pythonize_name('UPPER') -> 'upper'
        pythonize_name('ContentMd5')->'content_md5'
        pythonize_name('') -> ''
    """
    if name == "eTag":
        return "etag"
    s1 = _first_cap_regex.sub(r'\1_\2', name)
    s2 = _number_cap_regex.sub(r'\1_\2', s1)
    return _end_cap_regex.sub(r'\1_\2', s2).lower()


def get_canonical_querystring(params, for_signature):
    """

    :param params:
    :param for_signature:
    :return:
    """
    if params is None:
        return ''
    result = []
    for k, v in iteritems(params):
        if not for_signature or k.lower != http_headers.AUTHORIZATION.lower():
            if v is None:
                v = ''
            result.append(b'%s=%s' % (normalize_string(k), normalize_string(v)))
    result.sort()
    return (b'&').join(result)


def print_object(obj):
    """

    :param obj:
    :return:
    """
    tmp = []
    for k, v in iteritems(obj.__dict__):
        if not k.startswith('__'):
            if isinstance(v, bytes):
                tmp.append("%s:'%s'" % (k, v))
            # str is unicode
            elif isinstance(v, str):
                tmp.append("%s:u'%s'" % (k, v))
            else:
                tmp.append('%s:%s' % (k, v))
    return '{%s}' % ','.join(tmp)


class Expando(object):
    """
    Expandable class
    """
    def __init__(self, attr_dict=None):
        if attr_dict:
            self.__dict__.update(attr_dict)

    def __getattr__(self, item):
        if item.startswith('__'):
            raise AttributeError
        return None

    def __repr__(self):
        return print_object(self)


def dict_to_python_object(d):
    """

    :param d:
    :return:
    """
    attr = {}
    for k, v in iteritems(d):
        if not isinstance(k, compat.string_types):
            k = compat.convert_to_string(k)
        k = pythonize_name(k)
        attr[k] = v
    return Expando(attr)


def required(**types):
    """
    decorator of input param check
    :param types:
    :return:
    """
    def _required(f):
        def _decorated(*args, **kwds):
            for i, v in enumerate(args):
                if f.__code__.co_varnames[i] in types:
                    if v is None:
                        raise ValueError('arg "%s" should not be None' %
                                         (f.__code__.co_varnames[i]))
                    if not isinstance(v, types[f.__code__.co_varnames[i]]):
                        raise TypeError('arg "%s"= %r does not match %s' %
                                        (f.__code__.co_varnames[i],
                                         v,
                                         types[f.__code__.co_varnames[i]]))
            for k, v in iteritems(kwds):
                if k in types:
                    if v is None:
                        raise ValueError('arg "%s" should not be None' % k)
                    if not isinstance(v, types[k]):
                        raise TypeError('arg "%s"= %r does not match %s' % (k, v, types[k]))
            return f(*args, **kwds)
        _decorated.__name__ = f.__name__
        return _decorated
    return _required


def parse_host_port(endpoint, default_protocol):
    """
    parse protocol, host, port from endpoint in config

    :type: string
    :param endpoint: endpoint in config

    :type: baidubce.protocol.HTTP or baidubce.protocol.HTTPS
    :param default_protocol: if there is no scheme in endpoint,
                              we will use this protocol as default
    :return: tuple of protocol, host, port
    """
    # netloc should begin with // according to RFC1808
    if b"//" not in endpoint:
        endpoint = b"//" + endpoint

    try:
        # scheme in endpoint dominates input default_protocol
        parse_result = urlparse(
                endpoint,
                compat.convert_to_bytes(default_protocol.name))
    except Exception as e:
        raise ValueError('Invalid endpoint:%s, error:%s' % (endpoint,
            compat.convert_to_string(e)))

    if parse_result.scheme == compat.convert_to_bytes(baidubce.protocol.HTTP.name):
        protocol = baidubce.protocol.HTTP
        port = baidubce.protocol.HTTP.default_port
    elif parse_result.scheme == compat.convert_to_bytes(baidubce.protocol.HTTPS.name):
        protocol = baidubce.protocol.HTTPS
        port = baidubce.protocol.HTTPS.default_port
    else:
        raise ValueError('Unsupported protocol %s' % parse_result.scheme)
    host = parse_result.hostname
    if parse_result.port is not None:
        port = parse_result.port

    return protocol, host, port


def ComplexHandler(Obj):
    """
    The helper of serialization
    """
    if hasattr(Obj, 'jsonable'):
        return Obj.jsonable()
    else:
        raise ('Object of type %s with value of %s is not JSON serializable' % (type(Obj), repr(Obj)))

