################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
This module provides related methods for running agent
"""
from migrate_center_agent import utils
from migrate_center_agent import heart_beat
from migrate_center_agent import task_status
from migrate_center_agent import mount
from migrate_center_agent import boot
from migrate_center_agent import modify
from migrate_center_agent import umount
from migrate_center_agent.rsync import rsync
from baidubce import heartbeat_request
import migrate_center_agent

import os
import logging
import sys
import time
import datetime
import signal
import gevent
import psutil
import platform


LOG_FORMAT = "%(asctime)s - %(levelname)s - %(pathname)s - %(filename)s - %(funcName)s - %(lineno)s - %(message)s"
DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"
logging.basicConfig(filename="migrate-center.log", filemode="w",\
                    level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)


class Agent(object):
    """
    Class that runs the agent
    """
    def __init__(self):
        self.terminated = False
        self.migration_source_id = ""
        self.client = None
        self.rsync = None
        self.sync_finished = False
        self.heart_beat_request = None
        self.heart_beat_response = None
        self.mount = None
        self.boot = None
        self.modify = None
        self.umount = None
        self.total_data_size = 0
        self.synced_data_size = 0
        self.rsync_error = False
        self.user_cfg = utils.get_user_cfg()

        for sig in [signal.SIGINT, signal.SIGHUP, signal.SIGTERM, signal.SIGQUIT]:
            signal.signal(sig, self.terminate_signal_handler)

    def terminate_signal_handler(self, signum, frame):
        """
        Handle keyboard interrupt
        """
        logging.warning("The agent is forcibly terminated and the migration process will "
                        "exit, signum is: %s", signum)

    def register_migration_source(self, client):
        """
        Register migration source

        :param client: The agent client
        :type client: baidubce.agent_client.AgentClient
        """
        response = client.register_migration_source()
        if not response.success:
            logging.error("register source error: %s", response.err_msg)
            sys.exit(1)
        else:
            return response.migration_source_id

    def send_task_status(self):
        """
        Set the console state to error or syncFinish
        """
        self.client.heart_beat(self.heart_beat_request)

    def start_heart_beat(self, heart_beat):
        """
        Start send heart beat packet to console
        """
        while not self.terminated:
            heart_beat.heart_beat()
            gevent.sleep(migrate_center_agent.HEART_BEAT_INTERVAL)

    def mount_disks(self, heart_beat):
        """
        Mount system disk and data disk.
        """
        sub_task = heartbeat_request.SubTask(task_status.Step.MOUNTING.value)
        sub_task.progress = 0.0
        sub_task.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.heart_beat_request.task_info.add_sub_task(sub_task)
        self.mount = mount.Mount(self.heart_beat_response.task_metadata.dest_root_user,
                                 self.heart_beat_response.task_metadata.dest_ip,
                                 self.heart_beat_response.task_metadata.disk_to_trans)
        self.mount.heart_beat = heart_beat
        self.mount.mount()
        err, err_log, _ = self.mount.err, self.mount.err_log, self.mount.src_to_dest
        if err:
            self.heart_beat_request.task_info.status = task_status.Error
            self.heart_beat_request.task_info.err_log = err
            self.send_task_status()
            logging.error("Mount disks failed: %s", err_log)
            sys.exit(1)
        sub_task.progress = 1.0
        sub_task.end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

    def boot_disks(self, pf, heart_beat):
        """
        Install boot sector
        """
        sub_task = heartbeat_request.SubTask(task_status.Step.BOOTING.value)
        sub_task.progress = 0.0
        sub_task.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.heart_beat_request.task_info.add_sub_task(sub_task)
        self.boot = boot.Boot(self.heart_beat_response.task_metadata.dest_root_user,
                              self.heart_beat_response.task_metadata.dest_ip)
        self.boot.platform = pf
        self.boot.heart_beat = heart_beat
        self.boot.boot()
        err, err_log = self.boot.err, self.boot.err_log
        if err:
            self.heart_beat_request.task_info.status = task_status.Error
            self.heart_beat_request.task_info.err_log = err
            self.send_task_status()
            logging.error("Boot disks failed: %s", err_log)
            sys.exit(1)
        sub_task.progress = 1.0
        sub_task.end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

    def modify_uuid(self):
        """
        Replace the old uuid to the new uuid.
        """
        sub_task = heartbeat_request.SubTask(task_status.Step.MODIFYING.value)
        sub_task.progress = 0.0
        sub_task.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.heart_beat_request.task_info.add_sub_task(sub_task)
        self.modify = modify.Modify(self.heart_beat_response.task_metadata.dest_root_user,
                                    self.heart_beat_response.task_metadata.dest_ip)
        self.modify.replace_uuid()
        err, err_log = self.modify.err, self.modify.err_log
        if err:
            self.heart_beat_request.task_info.status = task_status.Error
            self.heart_beat_request.task_info.err_log = err
            self.send_task_status()
            logging.error("Modify UUID failed: %s", err_log)
            sys.exit(1)
        sub_task.progress = 1.0
        sub_task.end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

    def umount_disks(self):
        """
        Umount file system and systemn disk.
        """
        sub_task = heartbeat_request.SubTask(task_status.Step.UMOUNTING.value)
        sub_task.progress = 0.0
        sub_task.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.heart_beat_request.task_info.add_sub_task(sub_task)
        self.umount = umount.Umount(self.heart_beat_response.task_metadata.dest_root_user,
                                    self.heart_beat_response.task_metadata.dest_ip)
        self.umount.umount()
        err, err_log = self.umount.err, self.umount.err_log
        if err:
            self.heart_beat_request.task_info.status = task_status.Error
            self.heart_beat_request.task_info.err_log = err
            self.send_task_status()
            logging.error("Umount failed: %s", err_log)
            sys.exit(1)
        sub_task.progress = 1.0
        sub_task.end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

    def wait_cur_sync_task_finished(self, cur_disk_used_size, start_time, sync):
        """
        Wait until current sync task finished.

        :param cur_disk_used_size: Used space of the disk currently being synchronized
        :type cur_disk_used_size: int

        :param start_time: The start time of the current task
        :type start_time: datetime.datetime
        """
        rsync_process = psutil.Process(sync.process.pid)
        has_paused = False
        while not self.rsync.finished():
            cur_task_progress, speed = self.rsync.sync_execute_info
            self.synced_data_size += cur_disk_used_size * cur_task_progress
            total_progress = self.synced_data_size / self.total_data_size
            used_time = (datetime.datetime.now() - start_time).seconds
            if total_progress == 0 or sync.sync_execute_info[0] == 0:
                estimate_left_time = 999999
            else:
                estimate_left_time = int(used_time / sync.sync_execute_info[0] - used_time)
            self.heart_beat_request.task_info.set_speed_left_time_progress(
                speed, estimate_left_time, sync.sync_execute_info[0]
            )
            self.heart_beat_request.task_info.sub_task[-1].set_progress(sync.sync_execute_info[0])
            if not self.heart_beat_response.task_metadata.started and not has_paused:
                rsync_process.suspend()
                has_paused = True
                self.rsync.pause = True
            if self.heart_beat_response.task_metadata.started and has_paused:
                rsync_process.resume()
                has_paused = False
                self.rsync.pause = False
            gevent.sleep(2)

    def sync(self):
        """
        Sync files from local system disk and data disk.
        """
        sub_task = heartbeat_request.SubTask(task_status.Step.SYNCING.value)
        sub_task.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.heart_beat_request.task_info.add_sub_task(sub_task)
        self.total_data_size = utils.total_data_size(self.mount.src_to_dest.keys())
        task_metadata = self.heart_beat_response.get_task_metadata()
        dest_ip = task_metadata.dest_ip
        dest_root_user = task_metadata.dest_root_user
        bwlimit = task_metadata.bwlimit
        data_disk_mount_path = []
        for src_path in self.mount.src_to_dest:
            if src_path != "/":
                data_disk_mount_path.append(src_path)
        for src_path, dest_path in self.mount.src_to_dest.items():
            logging.info("src %s to dest: %s", src_path, dest_path)
            if src_path == "/":
                self.rsync = rsync.Rsync(src_path, dest_path, dest_ip, dest_root_user, bwlimit\
                                   , self.mount.data_disk_mount_point)
            else:
                self.rsync = rsync.Rsync(src_path, dest_path, dest_ip, dest_root_user, bwlimit)
            rsync_retry_times = 3 if "rsync_retry_times" not in self.user_cfg else self.user_cfg["rsync_retry_times"]
            rsync_retry_interval = 3 if "rsync_retry_interval" not in self.user_cfg else self.user_cfg["rsync_retry_interval"]
            for i in range(0, rsync_retry_times):
                self.rsync.start_sync()
                start_time = datetime.datetime.now()
                cur_disk_used_size = utils.disk_used_size(src_path)
                self.wait_cur_sync_task_finished(cur_disk_used_size, start_time, self.rsync)
                if self.rsync.process.returncode != 0:
                    logging.error("rsync error with code: %s", self.rsync.process.returncode)
                    if i == rsync_retry_times - 1:
                        logging.error("final rsync error after %s times", rsync_retry_times)
                        self.rsync_error = True
                        self.terminated = True
                        break
                    gevent.sleep(rsync_retry_interval)
                else:
                    break
        sub_task.progress = 1.0
        sub_task.end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        logging.info("sync finish, start_time: %s, end_time: %s", sub_task.start_time, sub_task.end_time)
        self.sync_finished = True

    def run(self, skip=False):
        """
        Run the agent
        """
        os.popen("export LANG=en_US.UTF-8")
        if not skip:
            process = os.popen('bash check_env.sh')
            output = process.readline().strip()
            if output == '1':
                logging.error("Please compile your kernel to install virtio")
                return
            elif output == '2':
                logging.error("Please make initramfs file")
                return
            elif output == '3':
                logging.error("Please check rsync version and make sure rsync >= 3.1.0")
                return

        pf = platform.platform().lower()

        self.client = utils.set_up_agent_client()
        self.migration_source_id = self.register_migration_source(self.client)
        self.heart_beat_request = heartbeat_request.HeartBeatRequest(migration_source_id=self.migration_source_id)
        self.heart_beat_response = heart_beat.HeartBeatResponse()
        hb = heart_beat.HeartBeat(heart_beat_request=self.heart_beat_request,
                                  heart_beat_response=self.heart_beat_response,
                                  agent_client=self.client)
        g1 = gevent.spawn(self.start_heart_beat, hb)
        g1.start()
        while not self.terminated:
            task_metadata = self.heart_beat_response.get_task_metadata()
            if not task_metadata.error and not self.rsync_error:
                if task_metadata.started:
                    self.heart_beat_request.task_info.status = task_status.Syncing
                    if self.mount is None:
                        self.mount_disks(hb)
                    if self.mount is not None and self.mount.finished:
                        self.sync()
                    if self.sync_finished:
                        self.boot_disks(pf, hb)
                    if self.boot is not None and self.boot.finished:
                        self.heart_beat_request.task_info.finished = True
                        self.send_task_status()
                        break
                    if self.rsync is not None and self.rsync.err:
                        logging.error("An error occurred while synchronizing files: %s", self.rsync.err_msg)
                        self.heart_beat_request.task_info.status = task_status.Error
                        self.send_task_status()
                        break
                else:
                    if self.rsync and self.rsync.is_running():
                        self.rsync.stop_sync()
            else:
                logging.error("The sync task is in error status, you should delete the task on the "
                              "console before you restart the agent on your machine ")
                break

            gevent.sleep(1)
        logging.info("self.terminated: %s", self.terminated)
