################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
This module provides methods for installing boot sector to the remote system disk.
"""

from migrate_center_agent import utils
import migrate_center_agent

import gevent
import logging
import shlex
import subprocess
class Boot(object):
    """
    Boot the system disk.
    """
    def __init__(self, root_user, ip):
        self.root_user = root_user
        self.ip = ip
        self.err = False
        self.err_log = None
        self.finished = False
        self.platform = ""
        self.heart_beat = None

    def execute_remote_command(self, command):
        """
        Execute command on remote server.
        """
        cmd = utils.ssh_command_prefix(self.root_user, self.ip) + "\"" + command + "\""
        cmd = shlex.split(cmd)
        logging.info("boot cmd: %s", cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            self.heart_beat.heart_beat()
            gevent.sleep(1)
        grub_sleep = 0
        while grub_sleep <= 60:
            self.heart_beat.heart_beat()
            grub_sleep += 1
            gevent.sleep(1)
        process.wait()
        err, err_log = utils.error_handler(process)
        return err, err_log

    def mount_proc(self):
        """
        Mount proc system to the system disk
        """
        command = "mount -t proc none {}/proc/".format(migrate_center_agent.SYSTEM_DISK_MOUNT_DIR)
        return self.execute_remote_command(command)

    def mount_bind(self):
        """
        Mount bind system to the system disk
        """
        command = "mount -o bind /dev/ {}/dev/".format(migrate_center_agent.SYSTEM_DISK_MOUNT_DIR)
        err, err_log = self.execute_remote_command(command)
        if err:
            return err, err_log
        command = "mount -o bind /sys/ {}/sys/".format(migrate_center_agent.SYSTEM_DISK_MOUNT_DIR)
        return self.execute_remote_command(command)

    def mount_devpts(self):
        """
        Mount devpts system to the system disk
        """
        command = "mount -t devpts devpts {}/dev/pts".format(migrate_center_agent.SYSTEM_DISK_MOUNT_DIR)
        return self.execute_remote_command(command)

    def grub_command(self):
        """
        Get the command used for installing grub into system disk
        """
        return utils.grub_install_executable() + " " + \
               utils.get_system_disk_from_mount_dir(self.root_user, self.ip)[:-1]

    def install_grub(self):
        """
        Install grub to the system disk
        """
        scp_cmd = "scp ./update_target.sh " + self.root_user + "@" + self.ip + ":/mnt/"
        process = subprocess.Popen(shlex.split(scp_cmd), stdout=subprocess.PIPE, universal_newlines=True)
        process.wait()
        scp_cmd = "scp ./install_target_grub.sh " + self.root_user + "@" + self.ip + ":/mnt/sysdisk"
        process = subprocess.Popen(shlex.split(scp_cmd), stdout=subprocess.PIPE, universal_newlines=True)
        process.wait()
        command = "bash -x /mnt/update_target.sh "
        if "centos" in self.platform:
            command += " -t CentOS"
        elif "ubuntu" in self.platform:
            command += " -t Ubuntu"
        elif "debian" in self.platform:
            command += " -t Debian"
        return self.execute_remote_command(command)

    def boot(self):
        """
        Install the boot sector to the system disk
        """
        logging.info("begin install grub")
        self.err, self.err_log = self.install_grub()
        logging.info("boot finish")
        self.finished = True
