################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
This module provides methods for mounting system disk and data disks to remote server
"""
from migrate_center_agent import utils
from baidubce import utils as bceutils
import migrate_center_agent
import shlex
import subprocess
import re
import logging
import gevent

class Mount(object):
    """
    Mount system disk and data disk to remote server.
    """
    def __init__(self, root_user, ip, disk_to_trans):
        self.root_user = root_user
        self.ip = ip
        self.disk_to_trans = disk_to_trans
        self.finished = False
        self.err = False
        self.err_log = None
        self.src_to_dest = {}
        self.data_disk_mount_point = []
        self.heart_beat = None

    def get_remote_disks_sizes_devices(self):
        """
        Get the devices name of the disks attached to the remote server and their sizes.
        """
        # make sure id_rsa permission right
        change_permission_cmd = shlex.split("chmod 0600 /root/.ssh/id_rsa")
        subprocess.Popen(change_permission_cmd, universal_newlines=True)
        i = 0
        while i < 260:
            i += 1
            gevent.sleep(1)
            self.heart_beat.heart_beat()
        self.heart_beat.heart_beat()
        cmd = utils.ssh_command_prefix(self.root_user, self.ip) + "\"fdisk -l\""
        cmd = shlex.split(cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            gevent.sleep(1)
        err, err_log = utils.error_handler(process)
        sizes_devices = []
        for log in iter(process.stdout.readline, ""):
            if log == '' and process.poll() is not None:
                break
            log = log.split()
            if len(log) > 1 and (log[0] == "Disk" or log[0] == "磁盘"):
                m = re.match(r'(/dev/[s,v]d[b-z])', log[1])
                if m is not None:
                    sizes_devices.append([int(log[4]), m.group()])
        self.heart_beat.heart_beat()

        return err, err_log, sizes_devices

    def create_partition(self, device):
        """
        Create partition on the attached disk.
        """
        self.heart_beat.heart_beat()
        cmd_prefix = shlex.split(utils.ssh_command_prefix(self.root_user, self.ip))
        tmp = ['printf "n\np\n1\n\n\nw\n" | fdisk ' + device]
        cmd = cmd_prefix + tmp
        logging.info("create partition cmd: %s", cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            self.heart_beat.heart_beat()
            gevent.sleep(1)
        err, err_log = utils.error_handler(process)
        self.heart_beat.heart_beat()
        return err, err_log

    def create_file_system(self, partition):
        """
        Create ext4 file system on the specified partition
        """
        self.heart_beat.heart_beat()
        cmd = utils.ssh_command_prefix(self.root_user, self.ip) + "\"mkfs -t ext4 {}\"".format(partition)
        cmd = shlex.split(cmd)
        logging.info("create file system cmd: %s", cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            self.heart_beat.heart_beat()
            gevent.sleep(1)
        err, err_log = utils.error_handler(process)
        self.heart_beat.heart_beat()
        return err, err_log

    def create_dest_dir(self, dest_dir):
        """
        Create the destination dir on the remote server
        """
        self.heart_beat.heart_beat()
        cmd = utils.ssh_command_prefix(self.root_user, self.ip) + "\"mkdir -p {}\"".format(dest_dir)
        cmd = shlex.split(cmd)
        logging.info("create dest dir cmd: %s", cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            self.heart_beat.heart_beat()
            gevent.sleep(1)
        self.heart_beat.heart_beat()

    def mount_partition(self, partition, dest_dir):
        """
        Mount the partition to the destination dir.
        """
        self.heart_beat.heart_beat()
        self.create_dest_dir(dest_dir)
        cmd = utils.ssh_command_prefix(self.root_user, self.ip) + "\"mount {} {}\"".format(partition, dest_dir)
        cmd = shlex.split(cmd)
        logging.info("mount partition cmd: %s", cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        while process.poll() is None:
            self.heart_beat.heart_beat()
            gevent.sleep(1)
        err, err_log = utils.error_handler(process)
        self.heart_beat.heart_beat()
        return err, err_log

    def get_source_data_disk_mount_path(self):
        self.heart_beat.heart_beat()
        partition_info = bceutils.get_partition_info()
        for i in range(len(partition_info)):
            if partition_info[i]["mount_point"] != "/":
                self.data_disk_mount_point.append(partition_info[i]["mount_point"])
        self.heart_beat.heart_beat()

    def mount(self):
        """
        Mount the disks to the remote server
        """
        target_root = ""
        target_root_index = 0
        for i in range(len(self.disk_to_trans)):
            if self.disk_to_trans[i][1] == "/":
                target_root_index = i
                target_root = self.disk_to_trans[i][2]

        logging.info("begin mount")
        err, err_log, remote_sizes_devices = self.get_remote_disks_sizes_devices()
        if err:
            logging.error("get remote disks size error: %s", err_log)
            self.err, self.err_log = err, err_log
            return
        remote_sizes_devices.sort()
        logging.info("remote_sizes_devices: %s", remote_sizes_devices)
        remote_partitions = []
        for _, device in remote_sizes_devices:
            logging.info("remote device: %s", device)
            err, err_log = self.create_partition(device)
            if err:
                # Device does not contain a recognized partition table log not matter
                logging.error("create partition error: %s", err_log)
            # sync root to target disk
            if device == target_root:
                self.create_file_system(device + "1")
                err, err_log = self.mount_partition(device + "1", migrate_center_agent.SYSTEM_DISK_MOUNT_DIR)
                if err:
                    logging.error("mount root partition error: %s, error_log: %s", err, err_log)
                    self.err, self.err_log = err, err_log
                    return
                self.src_to_dest["/"] = migrate_center_agent.SYSTEM_DISK_MOUNT_DIR
            else:
                remote_partitions.append(device + "1")
        for partition in remote_partitions:
            err, err_log = self.create_file_system(partition)
            if err:
                logging.error("create file system error: %s", err_log)
                self.err, self.err_log = err, err_log
                return
        local_sizes_mount_dirs = self.disk_to_trans
        local_sizes_mount_dirs.remove(local_sizes_mount_dirs[target_root_index])
        local_sizes_mount_dirs.sort()
        j = 1
        for i in range(len(remote_partitions)):
            # root disk has been rsynced before
            if local_sizes_mount_dirs[i][1] != "/":
                dest_mount_dir = migrate_center_agent.DATA_DISK_MOUNT_DIR_PREFIX + str(j)
                j += 1
                err, err_log = self.mount_partition(remote_partitions[i], dest_mount_dir)
                if err:
                    logging.error("mount disk partition error: %s", err_log)
                    self.err, self.err_log = err, err_log
                    return
                self.src_to_dest[local_sizes_mount_dirs[i][1]] = dest_mount_dir
        self.get_source_data_disk_mount_path()
        self.err, self.err_log = err, err_log
        self.finished = True
        logging.info("end mount")
