################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
This module provides methods related to synchronization
"""
from migrate_center_agent import rsync
from migrate_center_agent.rsync import utils

import logging
import subprocess
import shlex
import codecs
import gevent

codecs.register_error("strict", codecs.ignore_errors)

LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)

class Rsync(object):
    """
    This class encapsulates the execution of rsync
    """
    def __init__(self, src_path, dest_path, dest_ip, dest_root_user, bwlimit=0, data_disk_mount_path=None):
        self.dest_ip = dest_ip
        self.src_path = src_path
        self.dest_path = dest_path
        self.user = dest_root_user
        self.bwlimit = bwlimit
        # When synchronizing system disks, we need to exclude the mounting directory of the data disks.
        self.exclude_paths = utils.get_exclude_paths_from_cfg_files()
        logging.info("rsync data_disk_mount_path: %s", data_disk_mount_path)
        if data_disk_mount_path:
            if not self.exclude_paths:
                self.exclude_paths = data_disk_mount_path
            else:
                self.exclude_paths += data_disk_mount_path
        logging.info("exclude_paths: %s", self.exclude_paths)
        self.process = None
        self.return_code = None
        self.err = False
        self.err_msg = ""
        self.sync_execute_info = (0, 0)
        self.pause = False

    def commands(self):
        """
        Get rsync execute command

        :return
        :rtype list
        """
        prefix = rsync.RSYNC_CMD_PREFIX + self.src_path
        remote_dest = "{}@{}:{}".format(self.user, self.dest_ip, self.dest_path)
        cmd = prefix + " " + remote_dest
        if self.exclude_paths:
            for exclude_path in self.exclude_paths:
                cmd += " --exclude={} ".format(exclude_path)
        cmd += " --exclude-from={} ".format(rsync.DEFAULT_EXCLUDE_CFG_PATH)
        if self.bwlimit > 0:
            cmd += "--bwlimit={}".format(str(self.bwlimit))
        logging.info("rsync cmd: %s", cmd)
        return shlex.split(cmd)

    def start_sync(self):
        """
        Start execute rsync command
        """
        cmd = self.commands()
        self.execute(cmd)

    def stop_sync(self):
        """
        Stop rsync
        """
        if self.process is not None:
            self.process.terminate()
        logging.warning(
            "The rsync is exited and the sync progress is stopped, if you want to resume the sync "
            "progress, please do not exit the agent")

    def execute(self, cmd):
        """
        Execute the specified command in another process
        """
        #process = subprocess.Popen(cmd, universal_newlines=True)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        self.process = process
        g1 = gevent.spawn(self.update_sync_execute_info)
        g1.start()

    def update_sync_execute_info(self):
        """
        Update sync execute information
        """
        while self.process.poll() is None:
            if self.pause:
                logging.info("rsync has been stopped")
                gevent.sleep(1)
            else:
                stdout_line = self.process.stdout.readline()
                if stdout_line == '' and self.process.poll() is not None:
                    break
                logging.info(stdout_line)
                stdout_line = stdout_line.split()
                if len(stdout_line) >= 4:
                    try:
                        progress = int(stdout_line[1][:-1])
                        # prevent divisor from 0
                        if progress == 0:
                            progress = 0.1
                        progress = progress / 100
                        speed = stdout_line[2]
                        self.sync_execute_info = (progress, speed)
                    except Exception:
                        pass
                gevent.sleep(0.1)
        self.process.stdout.close()

    def get_err_msg(self):
        """
        Check for errors during synchronization
        """
        for log in iter(self.process.stderr.readline, ""):
            if log == '' and self.process.poll() is not None:
                break
            self.err_msg += log
        if self.process.poll() != 0:
            self.err = True

    def get_speed_progress_left_time(self):
        """
        Get the progress, speed, and left time of the transfer
        """
        return self.sync_execute_info

    def is_running(self):
        """
        Check if the rsync is running
        """
        if self.process is None:
            return False
        if self.process.poll() is None:
            return True
        return False

    def finished(self):
        """
        Check if the rsync process has finished
        """
        if self.process is None:
            return True
        if self.process.poll() is None:
            return False
        return True
