################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
This module provides related tools for running the agent
"""
import migrate_center_agent
from baidubce.auth.bce_credentials import BceCredentials
from baidubce.bce_client_configuration import BceClientConfiguration
from baidubce import agent_client

import json
import subprocess
import shlex
import psutil
import platform
import os
import logging

def get_ak_sk_host():
    """
    Read AK, SK, Host from user's input

    :return
    :rtype string, string, string
    """
    cfg_path = migrate_center_agent.USER_CFG_PATH
    with open(cfg_path, 'r') as cfg_file:
        user_cfg = json.load(cfg_file)
        host = user_cfg['host']
        ak = user_cfg['ak']
        sk = user_cfg['sk']

    return ak, sk, host


def set_up_agent_client():
    """
    Set up the agent client to interact with the console

    :return
    :rtype baidubce.agent_client.AgentClient
    """
    ak, sk, host = get_ak_sk_host()
    config = BceClientConfiguration(credentials=BceCredentials(ak, sk),
                                    endpoint=host)
    client = agent_client.AgentClient(config)
    return client


def get_local_disks_sizes_mount_dirs():
    """
    Get the size and mounting directory of the migration source disk
    """
    partitions = psutil.disk_partitions()
    sizes_mount_dirs = []
    for partition in partitions:
        size = psutil.disk_usage(partition.mountpoint).total
        sizes_mount_dirs.append([size, partition.mountpoint])
    return sizes_mount_dirs


def total_data_size(dirs):
    """
    Get the total size of all data that needs to be synchronized
    """
    total_size = 0
    for dir in dirs:
        total_size += psutil.disk_usage(dir).used
    return total_size


def get_user_cfg():
    """
    Read user config from user_cfg.json

    :return
    :rtype dict
    """
    cfg_path = migrate_center_agent.USER_CFG_PATH
    user_cfg = None
    with open(cfg_path, 'r') as cfg_file:
        user_cfg = json.load(cfg_file)

    return user_cfg


def disk_used_size(dir):
    """
    Get the used size of the specified disk
    """
    return psutil.disk_usage(dir).used


def ssh_command_prefix(user, ip):
    """
    Get the ssh command prefix, e.g. ssh user@ip
    """
    return "ssh -o \"StrictHostKeyChecking no\" -v " + user + "@" + ip + " "


def error_handler(process):
    """
    Check if error has happened
    """
    err = False
    err_log = None
    returncode = process.poll()

    if returncode != 0:
        logging.error("there is error return code: %s", returncode)
        err = True
        err_log = "".join(process.stderr.readlines())

    return err, err_log


def get_system_disk_from_mount_dir(user, ip):
    """
    Get the corresponding device name from the mount directory
    """
    cmd = ssh_command_prefix(user, ip) + "\"df\""
    cmd = shlex.split(cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    logs = process.stdout.readlines()
    device = None
    for log in logs:
        if migrate_center_agent.SYSTEM_DISK_MOUNT_DIR in log:
            device = log.split()[0]
    return device


def exist_command(name):
    """
    Check if a command exist.
    """
    return True


def grub_install_executable():
    """
    Get the grub install command, e.g, grub-install or grub2-install
    """
    possible_commands = ["grub-install", "grub2-install"]
    for command in possible_commands:
        if exist_command(command):
            return command


def boot_type():
    """
    Get the boot type of grub, e.g, BIOS or UEFI.
    """
    uefi_flag = "/sys/firmware/efi"
    if os.path.exists(uefi_flag):
        return migrate_center_agent.BOOT_UEFI
    else:
        return migrate_center_agent.BOOT_BIOS


def uefi_grub_cfg_path():
    """
    Get the grub config file path on uefi platform
    """
    os_dist = platform.dist()[0]
    return migrate_center_agent.UEFI_GRUB_CFG_PATH_PREFIX + os_dist.lower() + "/" + \
           migrate_center_agent.GRUB_CFG_FILE_NAME


def bios_grub_cfg_path():
    """
    Get the grub config file name on bios platform.
    """
    if os.path.exists(migrate_center_agent.BIOS_GRUB_CFG_PATH_PREFIX + "grub2"):
        return migrate_center_agent.BIOS_GRUB_CFG_PATH_PREFIX + "grub2/" + \
               migrate_center_agent.GRUB_CFG_FILE_NAME
    else:
        return migrate_center_agent.BIOS_GRUB_CFG_PATH_PREFIX + "grub/" + \
               migrate_center_agent.GRUB_CFG_FILE_NAME


def get_grub_cfg_path():
    """
    Get the grub config file path.
    """
    if boot_type() == migrate_center_agent.BOOT_BIOS:
        return bios_grub_cfg_path()
    else:
        return uefi_grub_cfg_path()
